def create_param_space(data, no_of_metrics=2):

    '''Takes as input experiment log dataframe and returns
    ParamSpace object

    data | DataFrame | Talos experiment log as pandas dataframe
    no_of_metrics | int | number of metrics in the dataframe

    '''

    from talos.parameters.ParamSpace import ParamSpace

    params = {}

    for col in data.iloc[:, no_of_metrics:].columns:
        params[col] = data[col].unique().tolist()

    param_keys = sorted(list(params.keys()))

    return ParamSpace(params, param_keys)



from keras.callbacks import Callback


class PowerDrawCallback(Callback):

    '''A callback for recording GPU power draw (watts) on epoch begin and end.

    Example use:

    power_draw = PowerDrawCallback()

    model.fit(...callbacks=[power_draw]...)

    print(power_draw.logs)

    '''

    def __init__(self):

        super(PowerDrawCallback, self).__init__()

        import os
        import time

        self.os = os
        self.time = time.time
        self.command = "nvidia-smi -i 0 -q | grep -i 'power draw' | tr -s ' ' | cut -d ' ' -f5"

    def on_train_begin(self, logs={}):
        self.log = {}
        self.log['epoch_begin'] = []
        self.log['epoch_end'] = []
        self.log['seconds'] = []

    def on_epoch_begin(self, batch, logs=None):
        self.epoch_start_time = self.time()
        temp = self.os.popen(self.command).read()
        temp = float(temp.strip())
        self.log['epoch_begin'].append(temp)

    def on_epoch_end(self, batch, logs=None):
        temp = self.os.popen(self.command).read()
        temp = float(temp.strip())
        self.log['epoch_end'].append(temp)
        seconds = round(self.time() - self.epoch_start_time, 3)
        self.log['seconds'].append(seconds)



class TorchHistory:

    '''This is a helper for replicating the history object
    behavior of Keras to make Talos Scan() API consistent between
    the two backends.'''

    def __init__(self):

        self.history = {}

    def init_history(self):
        self.history = {}

    def append_history(self, history_data, label):
        if label not in self.history.keys():
            self.history[label] = []
        self.history[label].append(history_data)

    def append_loss(self, _loss):
        self.append_history(_loss, 'loss')

    def append_metric(self, _metric):
        self.append_history(_metric, 'metric')

    def append_val_loss(self, _loss):
        self.append_history(_loss, 'val_loss')

    def append_val_metric(self, _loss):
        self.append_history(_loss, 'val_metric')



def rescale_meanzero(x):

    '''Rescales an array to mean-zero.

    x | array | the dataset to be rescaled
    '''

    import wrangle
    import pandas as pd

    return wrangle.df_rescale_meanzero(pd.DataFrame(x)).values



def power_draw_append(history, power_draw):

    '''For appending the data from PowerDrawCallback to the history object
    and allowing the data to be captured in the experiment log in Talos.'''

    import numpy as np

    joined = power_draw.log['epoch_begin'] + power_draw.log['epoch_end']
    avg_watts = (np.array(joined)) / 2

    history.history['watts_min'] = [min(joined)]
    history.history['watts_max'] = [max(joined)]
    history.history['seconds'] = [sum(power_draw.log['seconds'])]

    watt_seconds = round(sum(avg_watts * np.array(power_draw.log['seconds'])), 2)
    history.history['Ws'] = [watt_seconds]

    return history



def local_strategy(self):

    try:
        import importlib
        importlib.reload(talos_strategy)
        self = talos_strategy(self)
    except NameError:
        try:
            from talos_strategy import talos_strategy
            self = talos_strategy(self)
        except ImportError:
            print("No talos_strategy.py found in pwd. Nothing is done.")

    return self



def hidden_layers(model, params, last_neuron):
    '''HIDDEN LAYER Generator

    NOTE: 'shapes', 'first_neuron', 'dropout', and 'hidden_layers' need
    to be present in the params dictionary.

    Hidden layer generation for the cases where number
    of layers is used as a variable in the optimization process.
    Handles things in a way where any number of layers can be tried
    with matching hyperparameters.'''

    # check for the params that are required for hidden_layers

    from keras.layers import Dense, Dropout
    from .network_shape import network_shape
    from ..utils.exceptions import TalosParamsError

    required = ['shapes', 'first_neuron', 'dropout', 'hidden_layers', 'activation']
    for param in required:
        if param not in params:
            message = "hidden_layers requires '" + param + "' in params"
            raise TalosParamsError(message)

    layer_neurons = network_shape(params, last_neuron)

    for i in range(params['hidden_layers']):
        model.add(Dense(
            layer_neurons[i],
            kernel_initializer=params.get(
                'kernel_initializer',
                'glorot_uniform'
            ),
            kernel_regularizer=params.get('kernel_regularizer'),
            bias_initializer=params.get('bias_initializer', 'zeros'),
            bias_regularizer=params.get('bias_regularizer'),
            use_bias=params.get('use_bias', True),
            activity_regularizer=params.get('activity_regularizer'),
            kernel_constraint=params.get('kernel_constraint'),
            bias_constraint=params.get('bias_constraint'),
            activation=params.get('activation')
        ))
        model.add(Dropout(params['dropout']))



def output_layer(task, last_activation, y_train, y_val):

    import numpy as np

    # output layer
    if task == 'binary':
        activation = last_activation
        last_neuron = 1

    elif task == 'multi_class':
        activation = last_activation
        last_neuron = len(np.unique(np.hstack((y_train, y_val))))

    elif task == 'multi_label':
        activation = last_activation
        last_neuron = y_train.shape[1]

    elif task == 'continuous':
        activation = None
        last_neuron = 1

    return activation, last_neuron





def recover_best_model(x_train,
                       y_train,
                       x_val,
                       y_val,
                       experiment_log,
                       input_model,
                       x_cross=None,
                       y_cross=None,
                       n_models=5,
                       task='multi_label'):

    '''Recover best models from Talos experiment log.

    x_train | array | same as was used in the experiment
    y_train | array | same as was used in the experiment
    x_val | array | same as was used in the experiment
    y_val | array | same as was used in the experiment
    x_cross | array | data for the cross-validation or None for use x_val
    y_cross | array | data for the cross-validation or None for use y_val
    experiment_log | str | path to the Talos experiment log
    input_model | function | model used in the experiment
    n_models | int | number of models to cross-validate
    task | str | binary, multi_class, multi_label or continuous

    Returns a pandas dataframe with the cross-validation results
    and the models.

    '''

    import pandas as pd
    import sklearn as sk
    import numpy as np

    from talos.utils.validation_split import kfold

    # read the experiment log into a dataframe
    df = pd.read_csv(experiment_log)

    # handle input data scenarios

    if x_cross is None or y_cross is None:
        x_cross = x_val
        y_cross = y_val

    # for final output
    results = []
    models = []

    for i in range(n_models):

        # get the params for the model and train it
        params = df.sort_values('val_acc', ascending=False).drop('val_acc', 1).iloc[i].to_dict()
        history, model = input_model(x_train, y_train, x_val, y_val, params)

        # start kfold cross-validation
        out = []
        folds = 5
        kx, ky = kfold(x_cross, y_cross, folds, True)

        for i in range(folds):

            y_pred = model.predict(kx[i]).argmax(axis=1)

            if task == 'binary':
                y_pred = y_pred >= .5
                scores = sk.metrics.f1_score(y_pred, ky[i], average='binary')

            elif task == 'multi_class':
                y_pred = y_pred.argmax(axis=-1)
                scores = sk.metrics.f1_score(y_pred, ky[i], average='macro')

            if task == 'multi_label':
                y_pred = model.predict(kx[i]).argmax(axis=1)
                scores = sk.metrics.f1_score(y_pred,
                                             ky[i].argmax(axis=1),
                                             average='macro')

            elif task == 'continuous':
                y_pred = model.predict(kx[i])
                scores = sk.metrics.mean_absolute_error(y_pred, ky[i])

            out.append(scores)

        results.append(np.mean(out))
        models.append(model)

    out = df.sort_values('val_acc', ascending=False).head(n_models)
    out['crossval_mean_f1score'] = results

    return out, models
